import inspect
import asyncio
import concurrent.futures


# Copied from jupyterhub/utils.py
def maybe_future(obj):
    """Return an asyncio Future
    Use instead of gen.maybe_future
    For our compatibility, this must accept:
    - asyncio coroutine (gen.maybe_future doesn't work in tornado < 5)
    - tornado coroutine (asyncio.ensure_future doesn't work)
    - scalar (asyncio.ensure_future doesn't work)
    - concurrent.futures.Future (asyncio.ensure_future doesn't work)
    - tornado Future (works both ways)
    - asyncio Future (works both ways)
    """
    if inspect.isawaitable(obj):
        # already awaitable, use ensure_future
        return asyncio.ensure_future(obj)
    elif isinstance(obj, concurrent.futures.Future):
        return asyncio.wrap_future(obj)
    else:
        # could also check for tornado.concurrent.Future
        # but with tornado >= 5.1 tornado.Future is asyncio.Future
        f = asyncio.Future()
        f.set_result(obj)
        return f



from jupyter_server.utils import url_path_join

from .config import JupyterServerParameters
from .handlers import ParametersHandler


def _load_jupyter_server_extension(app):
    """
    This function is called when the extension is loaded.
    """
    app.web_app.settings["server_parameters"] = JupyterServerParameters(parent=app)

    handlers = [
        (
            url_path_join(app.web_app.settings["base_url"], "/parameters"),
            ParametersHandler,
        )
    ]
    app.web_app.add_handlers(".*$", handlers)


def _jupyter_server_extension_points():
    """
    Returns a list of dictionaries with metadata describing
    where to find the `_load_jupyter_server_extension` function.
    """
    return [{"module": "jupyter_server_parameters"}]



from jupyter_server.base.handlers import JupyterHandler
from urllib.parse import urljoin
import tornado.web


class ParametersHandler(JupyterHandler):
    @tornado.web.authenticated
    async def get(self, *args):
        params = {k: self.get_argument(k) for k in self.request.arguments}

        # Swallow out parameters URL
        redirect = params.pop("redirect", "/lab")

        # Send parameter values to server_params
        server_params = self.settings["server_parameters"]
        try:
            await server_params.handle_parameters(params)
        except Exception:
            self.log.exception("Unable to handle parameters")

        # Perform redirect
        self.redirect(urljoin(self.base_url, redirect))



from traitlets import default, Callable
from traitlets.config import LoggingConfigurable

from .utils import maybe_future


class JupyterServerParameters(LoggingConfigurable):
    handle_parameters_hook = Callable(
        help="""
        An optional hook function that you can implement to handle parameters 
        passed to the `/parameters` endpoint. This maybe a coroutine.
        Example::
            def my_hook(self, parameters):
                self.log.info(f"The secret key is {parameters['key']}")
            c.JupyterServerParameters.handle_parameters_hook = my_hook
        """
    ).tag(config=True)

    @default("handle_parameters_hook")
    def _handle_parameters_hook_default(self, _, params):
        self.log.debug(f"Received {params!r} parameters")

    async def handle_parameters(self, params):
        await maybe_future(self.handle_parameters_hook(self, params))
